@@ -534,6 +534,8 @@ impl<Exe: Executor> Connection<Exe> {
         auth_data: Option<Authentication>,
         proxy_to_broker_url: Option<String>,
         certificate_chain: &[Certificate],
+        allow_insecure_connection: bool,
+        tls_hostname_verification_enabled: bool,
         connection_timeout: Duration,
         operation_timeout: Duration,
         executor: Arc<Exe>,
@@ -592,6 +594,8 @@ impl<Exe: Executor> Connection<Exe> {
             auth_data,
             proxy_to_broker_url,
             certificate_chain,
+            allow_insecure_connection,
+            tls_hostname_verification_enabled,
             executor.clone(),
             operation_timeout,
         );
@@ -622,6 +626,8 @@ impl<Exe: Executor> Connection<Exe> {
         auth_data: Option<Authentication>,
         proxy_to_broker_url: Option<String>,
         certificate_chain: &[Certificate],
+        allow_insecure_connection: bool,
+        tls_hostname_verification_enabled: bool,
         executor: Arc<Exe>,
         operation_timeout: Duration,
     ) -> Result<ConnectionSender<Exe>, ConnectionError> {
@@ -635,6 +641,10 @@ impl<Exe: Executor> Connection<Exe> {
                     for certificate in certificate_chain {
                         builder.add_root_certificate(certificate.clone());
                     }
+                    builder.danger_accept_invalid_hostnames(
+                        allow_insecure_connection && !tls_hostname_verification_enabled,
+                    );
+                    builder.danger_accept_invalid_certs(allow_insecure_connection);
                     let cx = builder.build()?;
                     let cx = tokio_native_tls::TlsConnector::from(cx);
                     let stream = cx
@@ -677,6 +687,10 @@ impl<Exe: Executor> Connection<Exe> {
                     for certificate in certificate_chain {
                         connector = connector.add_root_certificate(certificate.clone());
                     }
+                    connector = connector.danger_accept_invalid_hostnames(
+                        allow_insecure_connection && !tls_hostname_verification_enabled,
+                    );
+                    connector = connector.danger_accept_invalid_certs(allow_insecure_connection);
                     let stream = connector
                         .connect(&hostname, stream)
                         .await
@@ -1089,7 +1103,7 @@ pub(crate) mod messages {
                             value: v.clone(),
                         })
                         .collect(),
-                    read_compacted: options.read_compacted,
+                    read_compacted: Some(options.read_compacted.unwrap_or(false)),
                     initial_position: Some(options.initial_position.into()),
                     schema: options.schema,
                     start_message_id: options.start_message_id,
