@@ -26,6 +26,7 @@ use crate::message::{
     BatchedMessage, Message as RawMessage, Metadata, Payload,
 };
 use crate::proto::{BaseCommand, CommandCloseConsumer};
+use crate::reader::{Reader, State};
 use crate::{BrokerAddress, DeserializeMessage, Pulsar};
 use core::iter;
 use rand::distributions::Alphanumeric;
@@ -61,6 +62,44 @@ pub struct ConsumerOptions {
     pub initial_position: InitialPosition,
 }
 
+impl ConsumerOptions {
+    /// within options, sets the priority level
+    pub fn with_priority_level(mut self, priority_level: i32) -> Self {
+        self.priority_level = Some(priority_level);
+        self
+    }
+
+    pub fn durable(mut self, durable: bool) -> Self {
+        self.durable = Some(durable);
+        self
+    }
+
+    pub fn starting_on_message(mut self, message_id_data: MessageIdData) -> Self {
+        self.start_message_id = Some(message_id_data);
+        self
+    }
+
+    pub fn with_metadata(mut self, metadata: BTreeMap<String, String>) -> Self {
+        self.metadata = metadata;
+        self
+    }
+
+    pub fn read_compacted(mut self, read_compacted: bool) -> Self {
+        self.read_compacted = Some(read_compacted);
+        self
+    }
+
+    pub fn with_schema(mut self, schema: Schema) -> Self {
+        self.schema = Some(schema);
+        self
+    }
+
+    pub fn with_initial_position(mut self, initial_position: InitialPosition) -> Self {
+        self.initial_position = initial_position;
+        self
+    }
+}
+
 #[derive(Debug, Clone)]
 pub struct DeadLetterPolicy {
     /// Maximum number of times that a message will be redelivered before being sent to the dead letter queue.
@@ -196,21 +235,16 @@ impl<T: DeserializeMessage, Exe: Executor> Consumer<T, Exe> {
                     try_join_all(topics.into_iter().map(|topic| client.lookup_topic(topic)))
                         .await?;
 
-                let topic_addr_pair: Vec<(String, BrokerAddress)> = c
-                    .topics
-                    .iter()
-                    .cloned()
-                    .zip(addrs.iter().cloned())
-                    .collect();
+                let topic_addr_pair = c.topics.iter().cloned().zip(addrs.iter().cloned());
 
-                let consumers = try_join_all(topic_addr_pair.into_iter().map(|(topic, addr)| {
+                let consumers = try_join_all(topic_addr_pair.map(|(topic, addr)| {
                     TopicConsumer::new(client.clone(), topic, addr, c.config().clone())
                 }))
                 .await?;
 
                 let consumers: BTreeMap<_, _> = consumers
                     .into_iter()
-                    .map(|c| (c.topic().to_owned(), Box::pin(c)))
+                    .map(|c| (c.topic(), Box::pin(c)))
                     .collect();
                 let topics = consumers.keys().cloned().collect();
                 let topic_refresh = Duration::from_secs(30);
@@ -240,7 +274,7 @@ impl<T: DeserializeMessage, Exe: Executor> Consumer<T, Exe> {
     /// returns the list of topics this consumer is subscribed on
     pub fn topics(&self) -> Vec<String> {
         match &self.inner {
-            InnerConsumer::Single(c) => vec![c.topic.clone()],
+            InnerConsumer::Single(c) => vec![c.topic()],
             InnerConsumer::Multi(c) => c.topics(),
         }
     }
@@ -372,15 +406,16 @@ enum InnerConsumer<T: DeserializeMessage, Exe: Executor> {
 
 type MessageIdDataReceiver = mpsc::Receiver<Result<(proto::MessageIdData, Payload), Error>>;
 
+// this is entirely public for use in reader.rs
 pub(crate) struct TopicConsumer<T: DeserializeMessage, Exe: Executor> {
-    consumer_id: u64,
-    config: ConsumerConfig,
+    pub(crate) consumer_id: u64,
+    pub(crate) config: ConsumerConfig,
     topic: String,
     messages: Pin<Box<MessageIdDataReceiver>>,
     engine_tx: mpsc::UnboundedSender<EngineMessage<Exe>>,
     #[allow(unused)]
     data_type: PhantomData<fn(Payload) -> T::Output>,
-    dead_letter_policy: Option<DeadLetterPolicy>,
+    pub(crate) dead_letter_policy: Option<DeadLetterPolicy>,
     last_message_received: Option<DateTime<Utc>>,
     messages_received: u64,
 }
@@ -570,11 +605,11 @@ impl<T: DeserializeMessage, Exe: Executor> TopicConsumer<T, Exe> {
         })
     }
 
-    fn topic(&self) -> &str {
-        &self.topic
+    pub fn topic(&self) -> String {
+        self.topic.clone()
     }
 
-    async fn connection(&mut self) -> Result<Arc<Connection<Exe>>, Error> {
+    pub async fn connection(&mut self) -> Result<Arc<Connection<Exe>>, Error> {
         let (resolver, response) = oneshot::channel();
         self.engine_tx
             .send(EngineMessage::GetConnection(resolver))
@@ -587,20 +622,24 @@ impl<T: DeserializeMessage, Exe: Executor> TopicConsumer<T, Exe> {
         })
     }
 
-    async fn check_connection(&mut self) -> Result<(), Error> {
+    pub async fn check_connection(&mut self) -> Result<(), Error> {
         let conn = self.connection().await?;
         info!("check connection for id {}", conn.id());
         conn.sender().send_ping().await?;
         Ok(())
     }
 
-    async fn ack(&mut self, msg: &Message<T>) -> Result<(), ConsumerError> {
+    pub async fn ack(&mut self, msg: &Message<T>) -> Result<(), ConsumerError> {
         self.engine_tx
             .send(EngineMessage::Ack(msg.message_id.clone(), false))
             .await?;
         Ok(())
     }
 
+    pub(crate) fn acker(&self) -> mpsc::UnboundedSender<EngineMessage<Exe>> {
+        self.engine_tx.clone()
+    }
+
     async fn cumulative_ack(&mut self, msg: &Message<T>) -> Result<(), ConsumerError> {
         self.engine_tx
             .send(EngineMessage::Ack(msg.message_id.clone(), true))
@@ -615,7 +654,7 @@ impl<T: DeserializeMessage, Exe: Executor> TopicConsumer<T, Exe> {
         Ok(())
     }
 
-    async fn seek(
+    pub async fn seek(
         &mut self,
         message_id: Option<MessageIdData>,
         timestamp: Option<u64>,
@@ -629,11 +668,11 @@ impl<T: DeserializeMessage, Exe: Executor> TopicConsumer<T, Exe> {
         Ok(())
     }
 
-    fn last_message_received(&self) -> Option<DateTime<Utc>> {
+    pub fn last_message_received(&self) -> Option<DateTime<Utc>> {
         self.last_message_received
     }
 
-    fn messages_received(&self) -> u64 {
+    pub fn messages_received(&self) -> u64 {
         self.messages_received
     }
 
@@ -691,7 +730,7 @@ struct ConsumerEngine<Exe: Executor> {
     _drop_signal: oneshot::Sender<()>,
 }
 
-enum EngineMessage<Exe: Executor> {
+pub(crate) enum EngineMessage<Exe: Executor> {
     Ack(MessageData, bool),
     Nack(MessageData),
     UnackedRedelivery,
@@ -884,7 +923,7 @@ impl<Exe: Executor> ConsumerEngine<Exe> {
                                     error!("could not ask for redelivery: {:?}", e);
                                 } else {
                                     for i in h.iter() {
-                                        self.unacked_messages.remove(&i);
+                                        self.unacked_messages.remove(i);
                                     }
                                 }
                             }
@@ -1012,14 +1051,11 @@ impl<Exe: Executor> ConsumerEngine<Exe> {
 
                 #[cfg(feature = "lz4")]
                 {
-                    use std::io::Read;
-
-                    let mut decompressed_payload = Vec::new();
-                    let mut decoder =
-                        lz4::Decoder::new(&payload.data[..]).map_err(ConsumerError::Io)?;
-                    decoder
-                        .read_to_end(&mut decompressed_payload)
-                        .map_err(ConsumerError::Io)?;
+                    let decompressed_payload = lz4::block::decompress(
+                        &payload.data[..],
+                        payload.metadata.uncompressed_size.map(|i| i as i32),
+                    )
+                    .map_err(ConsumerError::Io)?;
 
                     payload.data = decompressed_payload;
                     payload
@@ -1302,7 +1338,7 @@ impl Iterator for BatchedMessageIterator {
 
 /// Builder structure for consumers
 ///
-/// This is the main way to create a [Consumer]
+/// This is the main way to create a [Consumer] or a [Reader]
 #[derive(Clone)]
 pub struct ConsumerBuilder<Exe: Executor> {
     pulsar: Pulsar<Exe>,
@@ -1442,8 +1478,11 @@ impl<Exe: Executor> ConsumerBuilder<Exe> {
         self
     }
 
-    /// creates a [Consumer] from this builder
-    pub async fn build<T: DeserializeMessage>(self) -> Result<Consumer<T, Exe>, Error> {
+    // Checks the builder for inconsistencies
+    // returns a config and a list of topics with associated brokers
+    async fn validate<T: DeserializeMessage>(
+        self,
+    ) -> Result<(ConsumerConfig, Vec<(String, BrokerAddress)>), Error> {
         let ConsumerBuilder {
             pulsar,
             topics,
@@ -1451,15 +1490,23 @@ impl<Exe: Executor> ConsumerBuilder<Exe> {
             subscription,
             subscription_type,
             consumer_id,
-            consumer_name,
+            mut consumer_name,
             batch_size,
             unacked_message_resend_delay,
-            namespace,
-            topic_refresh,
             consumer_options,
             dead_letter_policy,
+            namespace: _,
+            topic_refresh: _,
         } = self;
 
+        if consumer_name.is_none() {
+            let s: String = (0..8)
+                .map(|_| rand::thread_rng().sample(Alphanumeric))
+                .map(|c| c as char)
+                .collect();
+            consumer_name = Some(format!("consumer_{}", s));
+        }
+
         if topics.is_none() && topic_regex.is_none() {
             return Err(Error::Custom(
                 "Cannot create consumer with no topics and no topic regex".into(),
@@ -1518,12 +1565,18 @@ impl<Exe: Executor> ConsumerBuilder<Exe> {
             options: consumer_options.unwrap_or_default(),
             dead_letter_policy,
         };
+        Ok((config, topics))
+    }
 
-        let consumers =
-            try_join_all(topics.into_iter().map(|(topic, addr)| {
-                TopicConsumer::new(pulsar.clone(), topic, addr, config.clone())
-            }))
-            .await?;
+    /// creates a [Consumer] from this builder
+    pub async fn build<T: DeserializeMessage>(self) -> Result<Consumer<T, Exe>, Error> {
+        // would this clone() consume too much memory?
+        let (config, joined_topics) = self.clone().validate::<T>().await?;
+
+        let consumers = try_join_all(joined_topics.into_iter().map(|(topic, addr)| {
+            TopicConsumer::new(self.pulsar.clone(), topic, addr, config.clone())
+        }))
+        .await?;
 
         let consumer = if consumers.len() == 1 {
             let consumer = consumers.into_iter().next().unwrap();
@@ -1531,15 +1584,19 @@ impl<Exe: Executor> ConsumerBuilder<Exe> {
         } else {
             let consumers: BTreeMap<_, _> = consumers
                 .into_iter()
-                .map(|c| (c.topic().to_owned(), Box::pin(c)))
+                .map(|c| (c.topic(), Box::pin(c)))
                 .collect();
             let topics = consumers.keys().cloned().collect();
-            let topic_refresh = topic_refresh.unwrap_or_else(|| Duration::from_secs(30));
-            let refresh = Box::pin(pulsar.executor.interval(topic_refresh).map(drop));
+            let topic_refresh = self
+                .topic_refresh
+                .unwrap_or_else(|| Duration::from_secs(30));
+            let refresh = Box::pin(self.pulsar.executor.interval(topic_refresh).map(drop));
             let mut consumer = MultiTopicConsumer {
-                namespace: namespace.unwrap_or_else(|| "public/default".to_string()),
-                topic_regex,
-                pulsar,
+                namespace: self
+                    .namespace
+                    .unwrap_or_else(|| "public/default".to_string()),
+                topic_regex: self.topic_regex,
+                pulsar: self.pulsar,
                 consumers,
                 topics,
                 new_consumers: None,
@@ -1557,29 +1614,54 @@ impl<Exe: Executor> ConsumerBuilder<Exe> {
         };
         Ok(Consumer { inner: consumer })
     }
+
+    /// creates a [Reader] from this builder
+    pub async fn into_reader<T: DeserializeMessage>(self) -> Result<Reader<T, Exe>, Error> {
+        // would this clone() consume too much memory?
+        let (mut config, mut joined_topics) = self.clone().validate::<T>().await?;
+
+        // the validate() function defaults sub_type to SubType::Shared,
+        // but a reader's subscription is exclusive
+        warn!("Subscription Type for a reader is `Exclusive`. Resetting.");
+        config.sub_type = SubType::Exclusive;
+
+        if self.topics.unwrap().len() > 1 {
+            return Err(Error::Custom(
+                "Unable to create a reader - one topic max".to_string(),
+            ));
+        }
+
+        let (topic, addr) = joined_topics.pop().unwrap();
+        let consumer = TopicConsumer::new(self.pulsar.clone(), topic, addr, config.clone()).await?;
+
+        Ok(Reader {
+            consumer,
+            state: Some(State::PollingConsumer),
+        })
+    }
 }
 
 /// the complete configuration of a consumer
 #[derive(Debug, Clone, Default)]
-pub struct ConsumerConfig {
+pub(crate) struct ConsumerConfig {
     /// subscription name
-    subscription: String,
+    pub(crate) subscription: String,
     /// subscription type
     ///
     /// default: Shared
-    sub_type: SubType,
+    pub(crate) sub_type: SubType,
     /// maximum size for batched messages
     ///
     /// default: 1000
-    batch_size: Option<u32>,
+    pub(crate) batch_size: Option<u32>,
     /// name of the consumer
-    consumer_name: Option<String>,
+    pub(crate) consumer_name: Option<String>,
     /// numerical id of the consumer
     consumer_id: Option<u64>,
     /// time after which unacked messages will be sent again
     unacked_message_redelivery_delay: Option<Duration>,
     /// consumer options
-    options: ConsumerOptions,
+    pub(crate) options: ConsumerOptions,
     /// dead letter policy
     dead_letter_policy: Option<DeadLetterPolicy>,
 }
@@ -1706,7 +1788,7 @@ impl<T: DeserializeMessage, Exe: Executor> MultiTopicConsumer<T, Exe> {
 
     async fn ack(&mut self, msg: &Message<T>) -> Result<(), ConsumerError> {
         if let Some(c) = self.consumers.get_mut(&msg.topic) {
-            c.ack(&msg).await
+            c.ack(msg).await
         } else {
             Err(ConnectionError::Unexpected(format!("no consumer for topic {}", msg.topic)).into())
         }
@@ -1714,7 +1796,7 @@ impl<T: DeserializeMessage, Exe: Executor> MultiTopicConsumer<T, Exe> {
 
     async fn cumulative_ack(&mut self, msg: &Message<T>) -> Result<(), ConsumerError> {
         if let Some(c) = self.consumers.get_mut(&msg.topic) {
-            c.cumulative_ack(&msg).await
+            c.cumulative_ack(msg).await
         } else {
             Err(ConnectionError::Unexpected(format!("no consumer for topic {}", msg.topic)).into())
         }
@@ -1722,7 +1804,7 @@ impl<T: DeserializeMessage, Exe: Executor> MultiTopicConsumer<T, Exe> {
 
     async fn nack(&mut self, msg: &Message<T>) -> Result<(), ConsumerError> {
         if let Some(c) = self.consumers.get_mut(&msg.topic) {
-            c.nack(&msg).await?;
+            c.nack(msg).await?;
             Ok(())
         } else {
             Err(ConnectionError::Unexpected(format!("no consumer for topic {}", msg.topic)).into())
@@ -1793,6 +1875,11 @@ impl<T> Message<T> {
     pub fn message_id(&self) -> &proto::MessageIdData {
         &self.message_id.id
     }
+
+    /// Get message key (partition key)
+    pub fn key(&self) -> Option<String> {
+        self.payload.metadata.partition_key.clone()
+    }
 }
 impl<T: DeserializeMessage> Message<T> {
     /// directly deserialize a message
